package com.capitalware.mqcm.mq;

import java.io.IOException;
import java.net.MalformedURLException;
import java.security.InvalidAlgorithmParameterException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Date;
import java.util.ResourceBundle;

import javax.swing.JOptionPane;

import com.ibm.mq.MQException;
import com.ibm.mq.MQQueue;
import com.ibm.mq.constants.CMQC;
import com.ibm.mq.constants.CMQCFC;
import com.ibm.mq.pcf.PCFException;
import com.ibm.mq.pcf.PCFMessage;
import com.ibm.mq.pcf.PCFMessageAgent;

import com.capitalware.mqcm.MQCM;
import com.capitalware.mqcm.handlers.ChannelNameFilter;
import com.capitalware.mqcm.objects.CWConstants;
import com.capitalware.mqcm.objects.MQObject;

/**
 * This class will handle all MQ interactions.
 *
 * @author Roger Lacroix
 * @version 1.2
 * @license Apache 2 License
 */
public class PCFChlStatus extends MyMQ
{
   private MQCM               parent;
   private boolean            working = true;
   private long               startTime;
   private long               endTime;
   private ArrayList<Object>  cells = null;
   private PCFMessageAgent    agent = null;
   private String             chlNameMask = "*";

   private ResourceBundle labels;

   /**
    * The constructor
    *
    * @param mqo MQObject
    * @param parent (MQChannelStatus)
    * @param labels
    */
   public PCFChlStatus(MQObject mqo, MQCM parent, ResourceBundle labels)
   {
      super(mqo);
      this.parent = parent;
      this.labels = labels;
   }

   /**
    * The run method is used to retrieve the MQ channel status
    * information in a separate thread.
    */
   public void run()
   {
      boolean majorFailure = false;
      String  popupTitle = "";
      String  popupMessage = "";

      try
      {
         connect();
         if (isCmdServerRunning())
            working = masterLoop();
         else
         {
            popupTitle = labels.getString("popup.msg.cmdserverisnotrunning");
            popupMessage = labels.getString("popup.msg.cmdserverneedstoberunning");
            majorFailure = true;
         }
      }
      catch (MQException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = labels.getString("lbl.field.completioncode") + " = " + e.completionCode + "\n" + labels.getString("lbl.field.reasoncode") + " = " + e.reasonCode;
         majorFailure = true;
      }
      catch (MalformedURLException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      catch (InvalidAlgorithmParameterException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      catch (NoSuchAlgorithmException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      catch (ClassNotFoundException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      catch (InstantiationException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      catch (IllegalAccessException e)
      {
         popupTitle = labels.getString("popup.error");
         popupMessage = e.getLocalizedMessage();
         majorFailure = true;
      }
      finally
      {
         try
         {
            if (_qMgr != null)
               _qMgr.disconnect();
         }
         catch (MQException e)
         {
            JOptionPane.showMessageDialog( parent,
                                           labels.getString("lbl.field.completioncode") + " = " + e.completionCode + "\n" + labels.getString("lbl.field.reasoncode") + " = " + e.reasonCode,
                                           labels.getString("popup.error"),
                                           JOptionPane.ERROR_MESSAGE);
         }
      }

      if (majorFailure)
      {
         JOptionPane.showMessageDialog( parent,
                                        popupMessage,
                                        popupTitle,
                                        JOptionPane.ERROR_MESSAGE);
         System.exit(99);
      }
   }

   /**
    * Check to see if the queue for the Command Server has been opened.
    * Hence, is the Command Server running.
    *
    * @return flag true / false
    * @throws MQException
    */
   private boolean isCmdServerRunning() throws MQException
   {
      boolean flag = false;
      int openInputOptions  = CMQC.MQOO_INQUIRE + CMQC.MQOO_FAIL_IF_QUIESCING;

      String cmdQName =  _qMgr.getCommandInputQueueName();

      MQQueue _cmdQ = _qMgr.accessQueue( cmdQName,
                                         openInputOptions,
                                         null,           // default q manager
                                         null,           // dynamic q name
                                         null );         // no alternate user id

      if (_cmdQ.getOpenInputCount() > 0)  /* the queue is open. */
         flag = true;

      _cmdQ.close();

      return flag;
   }

   /**
    * The masterLoop() method will loop endlessly sending MQ PCF commands
    * to the command server, parsing the results and then sleeping for 'x' seconds.
    * 'x' is user-defined by the RefreshRate XML tag.
    *
    * @return boolean
    */
   private boolean masterLoop()
   {
      boolean flag = true;
      startTime = timeIs();  // clock the original start time for a default value.

      try
      {
         agent = new PCFMessageAgent(_qMgr);

         while (working)
         {
            if (agent == null)
               reconnect();

            if (agent != null)
            {
               flag = handlePCFMessages();
               if (flag)
                  parent.updateTable(cells);
            }

            nowSleeping();
         }
      }
      catch (ArrayIndexOutOfBoundsException abe)
      {
         flag = false;
//         System.err.println("ArrayIndexOutOfBoundsException: Error: " + abe.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
						                      abe.getLocalizedMessage(),
                                        labels.getString("popup.error"),
						                      JOptionPane.ERROR_MESSAGE);
      }
      catch (NumberFormatException nfe)
      {
         flag = false;
//         System.err.println("NumberFormatException: Error: " + nfe.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
						                      nfe.getLocalizedMessage(),
                                        labels.getString("popup.error"),
						                      JOptionPane.ERROR_MESSAGE);
      }
      catch (PCFException pcfe)
      {
         flag = false;
         pcfErrorHandler(pcfe);
      }
      catch (MQException e)
      {
         flag = false;
//         System.err.println(e.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        labels.getString("lbl.field.completioncode") + " = " + e.completionCode + "\n" + labels.getString("lbl.field.reasoncode") + " = " + e.reasonCode,
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (Exception e)
      {
         flag = false;
//         System.err.println("Exception: Error: " + e.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }

      // Disconnect
      try
      {
         if (agent != null)
            agent.disconnect();
      }
      catch (MQException ioe)
      {}

      agent = null;
      return flag;
   }

   /**
    * Handle sending a PCF request message and receiving the PCF response messages.
    */
   private boolean handlePCFMessages()
   {
      boolean flag = true;
      cells = null;  //  don't need it any more! GC will notice it (hopefully!!)
      cells = new ArrayList<Object>();
      ChannelNameFilter cnf = new ChannelNameFilter(qm.monitorChlName);

      if (cnf.getFilterCount() == 1)
      {
          if (qm.monitorChlName.lastIndexOf("*") == -1 )
             chlNameMask = qm.monitorChlName;
          else if ( (qm.monitorChlName.length() - 1) == qm.monitorChlName.lastIndexOf("*") )
             chlNameMask = qm.monitorChlName;
         else
            chlNameMask = qm.monitorChlName.substring(0, qm.monitorChlName.lastIndexOf("*")+1);
      }

      try
      {
         PCFMessage   request;
         PCFMessage[] responses;

         /**
          * Build the PCF request
          */
         request = new PCFMessage(CMQCFC.MQCMD_INQUIRE_CHANNEL_STATUS);
         request.addParameter(CMQCFC.MQCACH_CHANNEL_NAME, chlNameMask);
         request.addParameter(CMQCFC.MQIACH_CHANNEL_INSTANCE_TYPE, CMQC.MQOT_CURRENT_CHANNEL);
         request.addParameter(CMQCFC.MQIACH_CHANNEL_INSTANCE_ATTRS,
                              new int []
                              {
                                 CMQCFC.MQCACH_CHANNEL_NAME,
                                 CMQCFC.MQCACH_CONNECTION_NAME,
                                 CMQCFC.MQIACH_CHANNEL_STATUS,
                                 CMQCFC.MQIACH_CHANNEL_SUBSTATE,
                                 CMQCFC.MQIACH_MSGS,
                                 CMQCFC.MQCACH_LAST_MSG_DATE,
                                 CMQCFC.MQCACH_LAST_MSG_TIME,
                                 CMQCFC.MQCACH_CHANNEL_START_DATE,
                                 CMQCFC.MQCACH_CHANNEL_START_TIME,
                                 CMQCFC.MQIACH_BYTES_SENT,
                                 CMQCFC.MQIACH_BYTES_RECEIVED,
                                 CMQCFC.MQIACH_BUFFERS_SENT,
                                 CMQCFC.MQIACH_BUFFERS_RECEIVED,
                                 CMQCFC.MQIACH_MCA_STATUS,
                                 CMQCFC.MQCACH_MCA_JOB_NAME,
                                 CMQCFC.MQCACH_MCA_USER_ID
                              } );

         responses = agent.send(request);

         // Don't care about errors when more than 2 structures are returned
         if ( (responses.length <= 2)  && (responses[0].getCompCode() != CMQC.MQCC_OK) )
            throw new PCFException( responses[0].getCompCode(), responses[0].getReason(), responses);

         Object[] nc = null;
         for (int i = 0; i < responses.length; i++)
         {
            // Make sure that each response is ok
            if ( ((responses[i]).getCompCode() == CMQC.MQCC_OK) &&
                 ((responses[i]).getParameterValue(CMQCFC.MQCACH_CHANNEL_NAME) != null) )
            {
               nc = parsePCFMessage(responses[i], cnf);

               if (nc != null)
                  cells.add(nc);
            }
         }

         responses = null;
         request = null;
      }
      catch (ArrayIndexOutOfBoundsException abe)
      {
         flag = false;
//         System.err.println("ArrayIndexOutOfBoundsException: Error: " + abe.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        abe.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (NumberFormatException nfe)
      {
         flag = false;
//         System.err.println("NumberFormatException: Error: " + nfe.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        nfe.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (PCFException pcfe)
      {
         flag = false;
         pcfErrorHandler(pcfe);
      }
      catch (MQException e)
      {
         flag = false;
//         System.err.println(e.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        labels.getString("lbl.field.completioncode") + " = " + e.completionCode + "\n" + labels.getString("lbl.field.reasoncode") + " = " + e.reasonCode,
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);

         reconnect();
      }
      catch (IOException ioe)
      {
         flag = false;
//         System.err.println("[1]IOException: Error: " + ioe.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        ioe.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (Exception e)
      {
         flag = false;
//         System.err.println("Exception: Error: " + e.getLocalizedMessage());
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }

      return flag;
   }

   /**
    * Parse the returned PCF response message
    * @param response
    * @param cnf
    * @return Object[]
    */
   private Object[] parsePCFMessage(PCFMessage   response, ChannelNameFilter cnf)
   {
      int xx = 0;
      int columns = CWConstants.columnWidth.length;
      Object[] _newCells = new Object[columns];
      for (int i=0; i < columns; i++)
         _newCells[i] = "";

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_CHANNEL_NAME).trim();

         if (!(cnf.accept((String)_newCells[0])))
         {
            _newCells = null;
            return _newCells;  // does not match - we're done.
         }
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_CONNECTION_NAME).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = getChannelStatus(response.getIntParameterValue(CMQCFC.MQIACH_CHANNEL_STATUS));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = getChannelSubState(response.getIntParameterValue(CMQCFC.MQIACH_CHANNEL_SUBSTATE));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = new Integer(response.getIntParameterValue(CMQCFC.MQIACH_MSGS));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_LAST_MSG_DATE).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_LAST_MSG_TIME).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = new Integer(response.getIntParameterValue(CMQCFC.MQIACH_BYTES_SENT));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = new Integer(response.getIntParameterValue(CMQCFC.MQIACH_BYTES_RECEIVED));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = new Integer(response.getIntParameterValue(CMQCFC.MQIACH_BUFFERS_SENT));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = new Integer(response.getIntParameterValue(CMQCFC.MQIACH_BUFFERS_RECEIVED));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_CHANNEL_START_DATE).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_CHANNEL_START_TIME).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = getMCAStatus(response.getIntParameterValue(CMQCFC.MQIACH_MCA_STATUS));
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_MCA_USER_ID).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }

      try
      {
         _newCells[xx++] = response.getStringParameterValue(CMQCFC.MQCACH_MCA_JOB_NAME).trim();
      }
      catch (PCFException e)
      {
         e.printStackTrace();
      }
      
      return _newCells;
   }

   /**
    * Attempt to reconnect to the QMgr
    */
   private void reconnect()
   {
      try
      {
         if (agent != null)
            agent.disconnect();
      }
      catch (MQException e)
      {}

      try
      {
         if (_qMgr != null)
            _qMgr.disconnect();
      }
      catch (MQException e)
      {}

      agent = null;
      _qMgr = null;

      try
      {
         Thread.sleep(30000); // time in milliseconds
      }
      catch (InterruptedException ie) {}

      try
      {
         connect();
         agent = new PCFMessageAgent(_qMgr);
      }
      catch (MQException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        labels.getString("lbl.field.completioncode") + " = " + e.completionCode + "\n" + labels.getString("lbl.field.reasoncode") + " = " + e.reasonCode,
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (MalformedURLException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (InvalidAlgorithmParameterException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (NoSuchAlgorithmException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (ClassNotFoundException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (InstantiationException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
      catch (IllegalAccessException e)
      {
         JOptionPane.showMessageDialog( parent,
                                        e.getLocalizedMessage(),
                                        labels.getString("popup.error"),
                                        JOptionPane.ERROR_MESSAGE);
      }
   }

   /**
    * Handle an error from a PCF command
    * @param pcfe
    * @param queueName
    */
   private void pcfErrorHandler(PCFException pcfe)
   {
      StringBuffer sb = new StringBuffer(24*1024);
      sb.append("PCFException: \n");
      sb.append(labels.getString("popup.error")+": '" + qm.monitorChlName + "'\n");
      boolean notFound = false;

      PCFMessage[]  responses2 = (PCFMessage []) pcfe.exceptionSource;

      for (int i = 0; i < responses2.length; i++)
      {
         sb.append(responses2[i] + "\n");

         if (responses2[i].getReason() == CMQCFC.MQRCCF_CHL_STATUS_NOT_FOUND)
            notFound = true;
      }

      if (notFound)
      {
         JOptionPane.showMessageDialog( parent,
                                        labels.getString("popup.msg.nochnlstatusfoundmatching") + " '" + chlNameMask + "'",
                                        labels.getString("popup.warning"),
                                        JOptionPane.WARNING_MESSAGE);
      }
      else
      {
//         System.err.println(sb.toString());
         JOptionPane.showMessageDialog( parent,
                                        sb.toString(),
                                        labels.getString("popup.error"),
          			                      JOptionPane.ERROR_MESSAGE);
      }
   }

   /**
    * Determine the channel status
    * @param chlStatus
    * @return chlStatus
    */
   private String getChannelStatus(int chlStatus)
   {
      String status = null;
      
      if (chlStatus == CMQCFC.MQCHS_INACTIVE)
         status = labels.getString("table.chl.status.inactive");
      else if (chlStatus == CMQCFC.MQCHS_BINDING)
         status = labels.getString("table.chl.status.binding");
      else if (chlStatus == CMQCFC.MQCHS_STARTING)
         status = labels.getString("table.chl.status.starting");
      else if (chlStatus == CMQCFC.MQCHS_RUNNING)
         status = labels.getString("table.chl.status.running");
      else if (chlStatus == CMQCFC.MQCHS_STOPPING)
         status = labels.getString("table.chl.status.stopping");
      else if (chlStatus == CMQCFC.MQCHS_RETRYING)
         status = labels.getString("table.chl.status.retrying");
      else if (chlStatus == CMQCFC.MQCHS_STOPPED)
         status = labels.getString("table.chl.status.stopped");
      else if (chlStatus == CMQCFC.MQCHS_REQUESTING)
         status = labels.getString("table.chl.status.requesting");
      else if (chlStatus == CMQCFC.MQCHS_PAUSED)
         status = labels.getString("table.chl.status.paused");
      else if (chlStatus == CMQCFC.MQCHS_DISCONNECTED)
         status = labels.getString("table.chl.status.disconnected");
      else if (chlStatus == CMQCFC.MQCHS_INITIALIZING)
         status = labels.getString("table.chl.status.initializing");
      else if (chlStatus == CMQCFC.MQCHS_SWITCHING)
         status = labels.getString("table.chl.status.switching");
      else
         status = labels.getString("table.cell.unknown") + " ["+chlStatus+"]";

      return status;
   }

   /**
    * Determine the channel SubStatus
    * @param chlSubState
    * @return subStatus
    */
   private String getChannelSubState(int chlSubState)
   {
      String subState = null;
      
      if (chlSubState == CMQCFC.MQCHSSTATE_OTHER)
         subState = labels.getString("table.chl.substate.other");
      else if (chlSubState == CMQCFC.MQCHSSTATE_END_OF_BATCH)
         subState = labels.getString("table.chl.substate.endofbatch");
      else if (chlSubState == CMQCFC.MQCHSSTATE_SENDING)
         subState = labels.getString("table.chl.substate.sending");
      else if (chlSubState == CMQCFC.MQCHSSTATE_RECEIVING)
         subState = labels.getString("table.chl.substate.receiving");
      else if (chlSubState == CMQCFC.MQCHSSTATE_SERIALIZING)
         subState = labels.getString("table.chl.substate.serializing");
      else if (chlSubState == CMQCFC.MQCHSSTATE_RESYNCHING)
         subState = labels.getString("table.chl.substate.resyncing");
      else if (chlSubState == CMQCFC.MQCHSSTATE_HEARTBEATING)
         subState = labels.getString("table.chl.substate.heartbeating");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_SCYEXIT)
         subState = labels.getString("table.chl.substate.insecurityexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_RCVEXIT)
         subState = labels.getString("table.chl.substate.inreceiveexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_SENDEXIT)
         subState = labels.getString("table.chl.substate.insendexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_MSGEXIT)
         subState = labels.getString("table.chl.substate.inmsgexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_MREXIT)
         subState = labels.getString("table.chl.substate.inmrexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_CHADEXIT)
         subState = labels.getString("table.chl.substate.inchadexit");
      else if (chlSubState == CMQCFC.MQCHSSTATE_NET_CONNECTING)
         subState = labels.getString("table.chl.substate.netconnecting");
      else if (chlSubState == CMQCFC.MQCHSSTATE_SSL_HANDSHAKING)
         subState = labels.getString("table.chl.substate.sslhandshaking");
      else if (chlSubState == CMQCFC.MQCHSSTATE_NAME_SERVER)
         subState = labels.getString("table.chl.substate.nameserver");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_MQPUT)
         subState = labels.getString("table.chl.substate.inmqput");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_MQGET)
         subState = labels.getString("table.chl.substate.inmqget");
      else if (chlSubState == CMQCFC.MQCHSSTATE_IN_MQI_CALL)
         subState = labels.getString("table.chl.substate.inmqicall");
      else if (chlSubState == CMQCFC.MQCHSSTATE_COMPRESSING)
         subState = labels.getString("table.chl.substate.compressing");
      else
         subState = labels.getString("table.cell.unknown") + " ["+chlSubState+"]";

      return subState;
   }

   /**
    * Determine the MCA Status
    * @param mcaStatus
    * @return status
    */
   private String getMCAStatus(int mcaStatus)
   {
      String status = null;
      
      if (mcaStatus == CMQCFC.MQMCAS_STOPPED)
         status = labels.getString("table.chl.status.stopped");
      else if (mcaStatus == CMQCFC.MQMCAS_RUNNING)
         status = labels.getString("table.chl.status.running");
      else
         status = labels.getString("table.cell.unknown") + " ["+mcaStatus+"]";

      return status;
   }

   /**
    * Get the current time.
    * @return long (timestamp)
    */
   private long timeIs()
   {
      Date currentDate = new java.util.Date();
      return currentDate.getTime();
   }

   /**
    * Have a little snooze.
    */
   private void nowSleeping()
   {
      long     waitTime;
      endTime = timeIs();  // clock the 'end' time

      waitTime = (qm.refreshRate * 1000L) - (endTime - startTime);
      if (waitTime <= 0 )
         waitTime = 20;  // make a default wait of 20 ms

      try
      {
         /*
          * Check for user input every 1/4 second (250ms)
          */
         while ( (waitTime > 0) && (working))
         {
            Thread.sleep(250); // time in milliseconds
            waitTime -= 250;
         }
      }
      catch (InterruptedException ie) {}

      startTime = timeIs();  // save the time because we are beginning the work again.
   }

   /**
    * Someone wants the MQ PCF collection to stop.
    */
   public void stopPlease()
   {
      working = false;
   }
}
